@@ -7,6 +7,14 @@
 #include <linux/slab.h>
 #include "apfs.h"
 
+struct apfs_node *apfs_query_root(const struct apfs_query *query)
+{
+	while (query->parent)
+		query = query->parent;
+	ASSERT(apfs_node_is_root(query->node));
+	return query->node;
+}
+
 static u64 apfs_catalog_base_oid(struct apfs_query *query)
 {
 	struct apfs_query *root_query = NULL;
@@ -162,21 +170,21 @@ static inline bool apfs_xid_in_snapshot(struct apfs_omap *omap, u64 xid)
 }
 
 /**
- * apfs_omap_lookup_block - Find the block number of a b-tree node from its id
+ * apfs_omap_lookup_block_with_xid - Find bno of a virtual object from oid/xid
  * @sb:		filesystem superblock
  * @omap:	object map to be searched
  * @id:		id of the node
+ * @xid:	transaction id
  * @block:	on return, the found block number
  * @write:	get write access to the object?
  *
- * Returns 0 on success or a negative error code in case of failure.
+ * Searches @omap for the most recent matching object with a transaction id
+ * below @xid. Returns 0 on success or a negative error code in case of failure.
  */
-int apfs_omap_lookup_block(struct super_block *sb, struct apfs_omap *omap,
-			   u64 id, u64 *block, bool write)
+static int apfs_omap_lookup_block_with_xid(struct super_block *sb, struct apfs_omap *omap, u64 id, u64 xid, u64 *block, bool write)
 {
 	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
 	struct apfs_query *query;
-	struct apfs_key key;
 	struct apfs_omap_map map = {0};
 	int ret = 0;
 
@@ -188,15 +196,13 @@ int apfs_omap_lookup_block(struct super_block *sb, struct apfs_omap *omap,
 	query = apfs_alloc_query(omap->omap_root, NULL /* parent */);
 	if (!query)
 		return -ENOMEM;
-
-	apfs_init_omap_key(id, apfs_mounted_xid(sb), &key);
-	query->key = &key;
+	apfs_init_omap_key(id, xid, &query->key);
 	query->flags |= APFS_QUERY_OMAP;
 
 	ret = apfs_btree_query(sb, &query);
 	if (ret) {
 		if (ret != -ENODATA)
-			apfs_err(sb, "query failed for oid 0x%llx, xid 0x%llx", id, apfs_mounted_xid(sb));
+			apfs_err(sb, "query failed for oid 0x%llx, xid 0x%llx", id, xid);
 		goto fail;
 	}
 
@@ -218,7 +224,7 @@ int apfs_omap_lookup_block(struct super_block *sb, struct apfs_omap *omap,
 
 		new_bh = apfs_read_object_block(sb, *block, write, preserve);
 		if (IS_ERR(new_bh)) {
-			apfs_err(sb, "CoW failed for oid 0x%llx, xid 0x%llx", id, apfs_mounted_xid(sb));
+			apfs_err(sb, "CoW failed for oid 0x%llx, xid 0x%llx", id, xid);
 			ret = PTR_ERR(new_bh);
 			goto fail;
 		}
@@ -234,7 +240,7 @@ int apfs_omap_lookup_block(struct super_block *sb, struct apfs_omap *omap,
 		else
 			ret = apfs_btree_replace(query, &key, sizeof(key), &val, sizeof(val));
 		if (ret)
-			apfs_err(sb, "CoW omap update failed (oid 0x%llx, xid 0x%llx)", id, apfs_mounted_xid(sb));
+			apfs_err(sb, "CoW omap update failed (oid 0x%llx, xid 0x%llx)", id, xid);
 
 		*block = new_bh->b_blocknr;
 		brelse(new_bh);
@@ -247,6 +253,36 @@ fail:
 	return ret;
 }
 
+/**
+ * apfs_omap_lookup_block - Find the block number of a b-tree node from its id
+ * @sb:		filesystem superblock
+ * @omap:	object map to be searched
+ * @id:		id of the node
+ * @block:	on return, the found block number
+ * @write:	get write access to the object?
+ *
+ * Returns 0 on success or a negative error code in case of failure.
+ */
+int apfs_omap_lookup_block(struct super_block *sb, struct apfs_omap *omap, u64 id, u64 *block, bool write)
+{
+	return apfs_omap_lookup_block_with_xid(sb, omap, id, apfs_mounted_xid(sb), block, write);
+}
+
+/**
+ * apfs_omap_lookup_newest_block - Find newest bno for a virtual object's oid
+ * @sb:		filesystem superblock
+ * @omap:	object map to be searched
+ * @id:		id of the object
+ * @block:	on return, the found block number
+ * @write:	get write access to the object?
+ *
+ * Returns 0 on success or a negative error code in case of failure.
+ */
+int apfs_omap_lookup_newest_block(struct super_block *sb, struct apfs_omap *omap, u64 id, u64 *block, bool write)
+{
+	return apfs_omap_lookup_block_with_xid(sb, omap, id, -1, block, write);
+}
+
 /**
  * apfs_create_omap_rec - Create a record in the volume's omap tree
  * @sb:		filesystem superblock
@@ -261,7 +297,6 @@ int apfs_create_omap_rec(struct super_block *sb, u64 oid, u64 bno)
 	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
 	struct apfs_omap *omap = sbi->s_omap;
 	struct apfs_query *query;
-	struct apfs_key key;
 	struct apfs_omap_key raw_key;
 	struct apfs_omap_val raw_val;
 	int ret;
@@ -269,9 +304,7 @@ int apfs_create_omap_rec(struct super_block *sb, u64 oid, u64 bno)
 	query = apfs_alloc_query(omap->omap_root, NULL /* parent */);
 	if (!query)
 		return -ENOMEM;
-
-	apfs_init_omap_key(oid, nxi->nx_xid, &key);
-	query->key = &key;
+	apfs_init_omap_key(oid, nxi->nx_xid, &query->key);
 	query->flags |= APFS_QUERY_OMAP;
 
 	ret = apfs_btree_query(sb, &query);
@@ -313,15 +346,12 @@ int apfs_delete_omap_rec(struct super_block *sb, u64 oid)
 	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
 	struct apfs_omap *omap = sbi->s_omap;
 	struct apfs_query *query;
-	struct apfs_key key;
 	int ret;
 
 	query = apfs_alloc_query(omap->omap_root, NULL /* parent */);
 	if (!query)
 		return -ENOMEM;
-
-	apfs_init_omap_key(oid, nxi->nx_xid, &key);
-	query->key = &key;
+	apfs_init_omap_key(oid, nxi->nx_xid, &query->key);
 	query->flags |= APFS_QUERY_OMAP;
 
 	ret = apfs_btree_query(sb, &query);
@@ -362,19 +392,28 @@ struct apfs_query *apfs_alloc_query(struct apfs_node *node,
 {
 	struct apfs_query *query;
 
-	query = kmalloc(sizeof(*query), GFP_KERNEL);
+	query = kzalloc(sizeof(*query), GFP_KERNEL);
 	if (!query)
 		return NULL;
 
 	/* To be released by free_query. */
 	query->node = node;
-	query->key = parent ? parent->key : NULL;
-	query->flags = parent ?
-		parent->flags & ~(APFS_QUERY_DONE | APFS_QUERY_NEXT) : 0;
-	query->parent = parent;
-	/* Start the search with the last record and go backwards */
-	query->index = node->records;
-	query->depth = parent ? parent->depth + 1 : 0;
+
+	if (parent) {
+		query->key = parent->key;
+		query->flags = parent->flags & ~(APFS_QUERY_DONE | APFS_QUERY_NEXT);
+		query->parent = parent;
+		query->depth = parent->depth + 1;
+	}
+
+	/*
+	 * We start the search with the last record and go backwards, but
+	 * some queries later use the PREV flag later to list them in order.
+	 */
+	if (query->flags & APFS_QUERY_PREV)
+		query->index = -1;
+	else
+		query->index = node->records;
 
 	return query;
 }
@@ -482,7 +521,8 @@ next_node:
 		 * than enough to map every block.
 		 */
 		apfs_err(sb, "btree is too high");
-		return -EFSCORRUPTED;
+		err = -EFSCORRUPTED;
+		goto fail;
 	}
 
 	err = apfs_node_query(sb, *query);
@@ -494,12 +534,16 @@ next_node:
 		err = apfs_query_set_before_first(sb, query);
 		if (err) {
 			apfs_err(sb, "failed to set before the first record");
-			return err;
+			goto fail;
 		}
-		return -ENODATA;
+		err = -ENODATA;
+		goto fail;
 	} else if (err == -EAGAIN) {
-		if (!(*query)->parent) /* We are at the root of the tree */
-			return -ENODATA;
+		if (!(*query)->parent) {
+			/* We are at the root of the tree */
+			err = -ENODATA;
+			goto fail;
+		}
 
 		/* Move back up one level and continue the query */
 		parent = (*query)->parent;
@@ -508,7 +552,7 @@ next_node:
 		*query = parent;
 		goto next_node;
 	} else if (err) {
-		return err;
+		goto fail;
 	}
 	if (apfs_node_is_leaf((*query)->node)) /* All done */
 		return 0;
@@ -517,14 +561,15 @@ next_node:
 	if (err) {
 		apfs_alert(sb, "bad index block: 0x%llx",
 			   (*query)->node->object.block_nr);
-		return err;
+		goto fail;
 	}
 
 	/* Now go a level deeper and search the child */
 	node = apfs_read_node(sb, child_id, storage, false /* write */);
 	if (IS_ERR(node)) {
 		apfs_err(sb, "failed to read node 0x%llx", child_id);
-		return PTR_ERR(node);
+		err = PTR_ERR(node);
+		goto fail;
 	}
 
 	if (node->object.oid != child_id)
@@ -539,12 +584,20 @@ next_node:
 	if (!*query) {
 		apfs_node_free(node);
 		*query = parent;
-		return -ENOMEM;
+		err = -ENOMEM;
+		goto fail;
 	}
 	node = NULL;
 	goto next_node;
+
+fail:
+	/* Don't leave stale record info here or some callers will use it */
+	(*query)->key_len = (*query)->len = 0;
+	return err;
 }
 
+static int __apfs_btree_replace(struct apfs_query *query, void *key, int key_len, void *val, int val_len);
+
 /**
  * apfs_query_join_transaction - Add the found node to the current transaction
  * @query: query that found the node
@@ -555,11 +608,22 @@ int apfs_query_join_transaction(struct apfs_query *query)
 	struct super_block *sb = node->object.sb;
 	u64 oid = node->object.oid;
 	u32 storage = apfs_query_storage(query);
+	struct apfs_obj_phys *raw = NULL;
+
+	/*
+	 * Ephemeral objects are checkpoint data, and all of their xids get
+	 * updated on commit. There is no real need to do it here as well, but
+	 * it's better for consistency with the other object types.
+	 */
+	if (storage == APFS_OBJ_EPHEMERAL) {
+		ASSERT(node->object.ephemeral);
+		raw = (void *)node->object.data;
+		raw->o_xid = cpu_to_le64(APFS_NXI(sb)->nx_xid);
+		return 0;
+	}
 
 	if (buffer_trans(node->object.o_bh)) /* Already in the transaction */
 		return 0;
-	/* Ephemeral objects are always checkpoint data */
-	ASSERT(storage != APFS_OBJ_EPHEMERAL);
 	/* Root nodes should join the transaction before the query is created */
 	ASSERT(!apfs_node_is_root(node));
 
@@ -575,9 +639,7 @@ int apfs_query_join_transaction(struct apfs_query *query)
 		__le64 bno = cpu_to_le64(node->object.block_nr);
 
 		/* The parent node needs to report the new location */
-		return apfs_btree_replace(query->parent,
-					  NULL /* key */, 0 /* key_len */,
-					  &bno, sizeof(bno));
+		return __apfs_btree_replace(query->parent, NULL /* key */, 0 /* key_len */, &bno, sizeof(bno));
 	}
 	return 0;
 }
@@ -604,9 +666,7 @@ static void apfs_btree_change_rec_count(struct apfs_query *query, int change,
 		ASSERT(!key_len && !val_len);
 	ASSERT(apfs_node_is_leaf(query->node));
 
-	while (query->parent)
-		query = query->parent;
-	root = query->node;
+	root = apfs_query_root(query);
 	ASSERT(apfs_node_is_root(root));
 
 	sb = root->object.sb;
@@ -637,11 +697,7 @@ void apfs_btree_change_node_count(struct apfs_query *query, int change)
 	struct apfs_btree_node_phys *root_raw;
 	struct apfs_btree_info *info;
 
-	ASSERT(!apfs_node_is_leaf(query->node));
-
-	while (query->parent)
-		query = query->parent;
-	root = query->node;
+	root = apfs_query_root(query);
 	ASSERT(apfs_node_is_root(root));
 
 	sb = root->object.sb;
@@ -654,91 +710,145 @@ void apfs_btree_change_node_count(struct apfs_query *query, int change)
 
 /**
  * apfs_query_refresh - Recreate a catalog query invalidated by node splits
- * @old_query: the catalog query to refresh
+ * @old_query:	query chain to refresh
+ * @root:	root node of the query chain
+ * @nodata:	is the query expected to find nothing?
  *
  * On success, @old_query is left pointing to the same leaf record, but with
  * valid ancestor queries as well. Returns a negative error code in case of
  * failure, or 0 on success.
  */
-static int apfs_query_refresh(struct apfs_query *old_query)
+static int apfs_query_refresh(struct apfs_query *old_query, struct apfs_node *root, bool nodata)
 {
-	struct apfs_node *node = old_query->node;
-	struct super_block *sb = node->object.sb;
-	char *raw = node->object.data;
-	struct apfs_query *new_query, *ancestor;
-	struct apfs_key new_key;
-	bool hashed = apfs_is_normalization_insensitive(sb);
+	struct super_block *sb = NULL;
+	struct apfs_query *new_query = NULL;
 	int err = 0;
 
-	/*
-	 * This function is for handling multiple splits of the same node,
-	 * which are only expected when large inline xattr values are involved.
-	 */
-	if ((old_query->flags & APFS_QUERY_TREE_MASK) != APFS_QUERY_CAT) {
-		apfs_err(sb, "non-catalog query");
+	sb = root->object.sb;
+
+	if (!apfs_node_is_leaf(old_query->node)) {
+		apfs_alert(sb, "attempting refresh of non-leaf query");
 		return -EFSCORRUPTED;
 	}
-	if (!apfs_node_is_leaf(node)) {
-		apfs_err(sb, "non-leaf query");
+	if (apfs_node_is_root(old_query->node)) {
+		apfs_alert(sb, "attempting refresh of root query");
 		return -EFSCORRUPTED;
 	}
 
-	/* Build a new query that points exactly to the same key */
-	err = apfs_read_cat_key(raw + old_query->key_off, old_query->key_len, &new_key, hashed);
-	if (err) {
-		apfs_err(sb, "failed to read the key");
-		return err;
-	}
-	new_query = apfs_alloc_query(APFS_SB(sb)->s_cat_root, NULL /* parent */);
+	new_query = apfs_alloc_query(root, NULL /* parent */);
 	if (!new_query)
 		return -ENOMEM;
-	new_query->key = &new_key;
-	new_query->flags = APFS_QUERY_CAT | APFS_QUERY_EXACT;
+	new_query->key = old_query->key;
+	new_query->flags = old_query->flags & ~(APFS_QUERY_DONE | APFS_QUERY_NEXT);
 
 	err = apfs_btree_query(sb, &new_query);
-	if (err) {
+	if (!nodata && err == -ENODATA) {
+		apfs_err(sb, "record should exist");
+		err = -EFSCORRUPTED;
+		goto fail;
+	}
+	if (err && err != -ENODATA) {
 		apfs_err(sb, "failed to rerun");
 		goto fail;
 	}
-
-	/* Set the original query flags and key on the new query */
-	for (ancestor = new_query; ancestor; ancestor = ancestor->parent) {
-		ancestor->flags = old_query->flags;
-		ancestor->key = old_query->key;
-	}
+	err = 0;
 
 	/* Replace the parent of the original query with the new valid one */
 	apfs_free_query(old_query->parent);
 	old_query->parent = new_query->parent;
 	new_query->parent = NULL;
 
+	/*
+	 * The records may have moved around so update this too. TODO: rework
+	 * the query struct so this stuff is not needed.
+	 */
+	ASSERT(old_query->node->object.oid == new_query->node->object.oid);
+	old_query->index = new_query->index;
+	old_query->key_off = new_query->key_off;
+	old_query->key_len = new_query->key_len;
+	old_query->off = new_query->off;
+	old_query->len = new_query->len;
+	old_query->depth = new_query->depth;
+
 fail:
 	apfs_free_query(new_query);
 	return err;
 }
 
 /**
- * apfs_query_is_orphan - Check if all of a query's ancestors are set
- * @query: the query to check
+ * __apfs_btree_insert - Insert a new record into a b-tree (at any level)
+ * @query:	query run to search for the record
+ * @key:	on-disk record key
+ * @key_len:	length of @key
+ * @val:	on-disk record value (NULL for ghost records)
+ * @val_len:	length of @val (0 for ghost records)
  *
- * A query may lose some of its ancestors during a node split. This can be
- * used to check if that has happened.
- *
- * TODO: running this check early on the insert, remove and replace functions
- * could be used to simplify several callers that do their own query refresh.
+ * The new record is placed right after the one found by @query.  On success,
+ * returns 0 and sets @query to the new record; returns a negative error code
+ * in case of failure, which may be -EAGAIN if a split happened and the caller
+ * must retry.
  */
-static bool apfs_query_is_orphan(const struct apfs_query *query)
+int __apfs_btree_insert(struct apfs_query *query, void *key, int key_len, void *val, int val_len)
 {
-	while (query) {
-		if (apfs_node_is_root(query->node))
-			return false;
-		query = query->parent;
+	struct apfs_node *node = query->node;
+	struct super_block *sb = node->object.sb;
+	struct apfs_btree_node_phys *node_raw;
+	int needed_room;
+	int err;
+
+	apfs_assert_query_is_valid(query);
+
+	err = apfs_query_join_transaction(query);
+	if (err) {
+		apfs_err(sb, "query join failed");
+		return err;
 	}
-	return true;
+
+	node = query->node;
+	node_raw = (void *)node->object.data;
+	apfs_assert_in_transaction(node->object.sb, &node_raw->btn_o);
+
+	needed_room = key_len + val_len;
+	if (!apfs_node_has_room(node, needed_room, false /* replace */)) {
+		if (node->records == 1) {
+			/* The new record just won't fit in the node */
+			err = apfs_create_single_rec_node(query, key, key_len, val, val_len);
+			if (err && err != -EAGAIN)
+				apfs_err(sb, "failed to create single-record node");
+			return err;
+		}
+		err = apfs_node_split(query);
+		if (err && err != -EAGAIN) {
+			apfs_err(sb, "node split failed");
+			return err;
+		}
+		return -EAGAIN;
+	}
+
+	apfs_assert_query_is_valid(query);
+
+	if (query->parent && query->index == -1) {
+		/* We are about to insert a record before all others */
+		err = __apfs_btree_replace(query->parent, key, key_len, NULL /* val */, 0 /* val_len */);
+		if (err) {
+			if (err != -EAGAIN)
+				apfs_err(sb, "parent update failed");
+			return err;
+		}
+	}
+
+	apfs_assert_query_is_valid(query);
+
+	err = apfs_node_insert(query, key, key_len, val, val_len);
+	if (err) {
+		apfs_err(sb, "node record insertion failed");
+		return err;
+	}
+	return 0;
 }
 
 /**
- * apfs_btree_insert - Insert a new record into a b-tree
+ * apfs_btree_insert - Insert a new record into a b-tree leaf
  * @query:	query run to search for the record
  * @key:	on-disk record key
  * @key_len:	length of @key
@@ -749,71 +859,45 @@ static bool apfs_query_is_orphan(const struct apfs_query *query)
  * returns 0 and sets @query to the new record; returns a negative error code
  * in case of failure.
  */
-int apfs_btree_insert(struct apfs_query *query, void *key, int key_len,
-		      void *val, int val_len)
+int apfs_btree_insert(struct apfs_query *query, void *key, int key_len, void *val, int val_len)
 {
-	struct apfs_node *node = query->node;
-	struct super_block *sb = node->object.sb;
-	struct apfs_btree_node_phys *node_raw;
+	struct super_block *sb = NULL;
+	struct apfs_node *root = NULL, *leaf = NULL;
 	int err;
 
-	/* Do this first, or node splits may cause @query->parent to be gone */
-	if (apfs_node_is_leaf(node))
-		apfs_btree_change_rec_count(query, 1 /* change */,
-					    key_len, val_len);
+	root = apfs_query_root(query);
+	ASSERT(apfs_node_is_root(root));
+	leaf = query->node;
+	ASSERT(apfs_node_is_leaf(leaf));
+	sb = root->object.sb;
 
-	err = apfs_query_join_transaction(query);
-	if (err) {
-		apfs_err(sb, "query join failed");
-		return err;
-	}
-
-again:
-	node = query->node;
-	node_raw = (void *)node->object.data;
-	apfs_assert_in_transaction(node->object.sb, &node_raw->btn_o);
-
-	err = apfs_node_insert(query, key, key_len, val, val_len);
-	if (err == -ENOSPC) {
-		if (!query->parent && !apfs_node_is_root(node)) {
-			err = apfs_query_refresh(query);
-			if (err) {
-				apfs_err(sb, "query refresh failed");
+	while (true) {
+		err = __apfs_btree_insert(query, key, key_len, val, val_len);
+		if (err != -EAGAIN) {
+			if (err)
 				return err;
-			}
-			if (node->records == 1) {
-				/* The new record just won't fit in the node */
-				return apfs_create_single_rec_node(query, key, key_len, val, val_len);
-			}
+			break;
 		}
-		err = apfs_node_split(query);
+		err = apfs_query_refresh(query, root, true /* nodata */);
 		if (err) {
-			apfs_err(sb, "node split failed");
+			apfs_err(sb, "query refresh failed");
 			return err;
 		}
-		goto again;
-	} else if (err) {
-		apfs_err(sb, "node record insertion failed");
-		return err;
 	}
 
-	/* This can only happen when we insert a record before all others */
-	if (query->parent && query->index == 0) {
-		err = apfs_btree_replace(query->parent, key, key_len,
-					 NULL /* val */, 0 /* val_len */);
-		if (err)
-			apfs_err(sb, "parent update failed");
-	}
-	return err;
+	apfs_assert_query_is_valid(query);
+	apfs_btree_change_rec_count(query, 1 /* change */, key_len, val_len);
+	return 0;
 }
 
 /**
- * apfs_btree_remove - Remove a record from a b-tree
+ * __apfs_btree_remove - Remove a record from a b-tree (at any level)
  * @query:	exact query that found the record
  *
- * Returns 0 on success, or a negative error code in case of failure.
+ * Returns 0 on success, or a negative error code in case of failure, which may
+ * be -EAGAIN if a split happened and the caller must retry.
  */
-int apfs_btree_remove(struct apfs_query *query)
+static int __apfs_btree_remove(struct apfs_query *query)
 {
 	struct apfs_node *node = query->node;
 	struct super_block *sb = node->object.sb;
@@ -821,12 +905,7 @@ int apfs_btree_remove(struct apfs_query *query)
 	int later_entries = node->records - query->index - 1;
 	int err;
 
-	/* Do this first, or node splits may cause @query->parent to be gone */
-	if (apfs_node_is_leaf(node))
-		apfs_btree_change_rec_count(query, -1 /* change */,
-					    0 /* key_len */, 0 /* val_len */);
-	else
-		apfs_btree_change_node_count(query, -1 /* change */);
+	apfs_assert_query_is_valid(query);
 
 	err = apfs_query_join_transaction(query);
 	if (err) {
@@ -838,17 +917,26 @@ int apfs_btree_remove(struct apfs_query *query)
 	node_raw = (void *)query->node->object.data;
 	apfs_assert_in_transaction(node->object.sb, &node_raw->btn_o);
 
-	if (node->records == 1) {
-		if (query->parent) {
-			/* Just get rid of the node */
-			return apfs_delete_node(query);
-		} else {
-			/* All descendants are gone, root is the whole tree */
-			node_raw->btn_level = 0;
-			node->flags |= APFS_BTNODE_LEAF;
+	if (query->parent && node->records == 1) {
+		/* Just get rid of the node */
+		err = __apfs_btree_remove(query->parent);
+		if (err == -EAGAIN)
+			return -EAGAIN;
+		if (err) {
+			apfs_err(sb, "parent index removal failed");
+			return err;
 		}
+		apfs_btree_change_node_count(query, -1 /* change */);
+		err = apfs_delete_node(node, query->flags & APFS_QUERY_TREE_MASK);
+		if (err) {
+			apfs_err(sb, "node deletion failed");
+			return err;
+		}
+		return 0;
 	}
 
+	apfs_assert_query_is_valid(query);
+
 	/* The first key in a node must match the parent record's */
 	if (query->parent && query->index == 0) {
 		int first_key_len, first_key_off;
@@ -859,14 +947,17 @@ int apfs_btree_remove(struct apfs_query *query)
 			return -EFSCORRUPTED;
 		key = (void *)node_raw + first_key_off;
 
-		err = apfs_btree_replace(query->parent, key, first_key_len,
-					 NULL /* val */, 0 /* val_len */);
+		err = __apfs_btree_replace(query->parent, key, first_key_len, NULL /* val */, 0 /* val_len */);
+		if (err == -EAGAIN)
+			return -EAGAIN;
 		if (err) {
 			apfs_err(sb, "parent update failed");
 			return err;
 		}
 	}
 
+	apfs_assert_query_is_valid(query);
+
 	/* Remove the entry from the table of contents */
 	if (apfs_node_has_fixed_kv_size(node)) {
 		struct apfs_kvoff *toc_entry;
@@ -888,6 +979,11 @@ int apfs_btree_remove(struct apfs_query *query)
 	apfs_node_free_range(node, query->off, query->len);
 
 	--node->records;
+	if (node->records == 0) {
+		/* All descendants are gone, root is the whole tree */
+		node_raw->btn_level = 0;
+		node->flags |= APFS_BTNODE_LEAF;
+	}
 	apfs_update_node(node);
 
 	--query->index;
@@ -895,7 +991,123 @@ int apfs_btree_remove(struct apfs_query *query)
 }
 
 /**
- * apfs_btree_replace - Replace a record in a b-tree
+ * apfs_btree_remove - Remove a record from a b-tree leaf
+ * @query:	exact query that found the record
+ *
+ * Returns 0 on success, or a negative error code in case of failure.
+ */
+int apfs_btree_remove(struct apfs_query *query)
+{
+	struct super_block *sb = NULL;
+	struct apfs_node *root = NULL, *leaf = NULL;
+	int err;
+
+	root = apfs_query_root(query);
+	ASSERT(apfs_node_is_root(root));
+	leaf = query->node;
+	ASSERT(apfs_node_is_leaf(leaf));
+	sb = root->object.sb;
+
+	while (true) {
+		err = __apfs_btree_remove(query);
+		if (err != -EAGAIN) {
+			if (err)
+				return err;
+			break;
+		}
+		err = apfs_query_refresh(query, root, false /* nodata */);
+		if (err) {
+			apfs_err(sb, "query refresh failed");
+			return err;
+		}
+	}
+
+	apfs_assert_query_is_valid(query);
+	apfs_btree_change_rec_count(query, -1 /* change */, 0 /* key_len */, 0 /* val_len */);
+	return 0;
+}
+
+/**
+ * __apfs_btree_replace - Replace a record in a b-tree (at any level)
+ * @query:	exact query that found the record
+ * @key:	new on-disk record key (NULL if unchanged)
+ * @key_len:	length of @key
+ * @val:	new on-disk record value (NULL if unchanged)
+ * @val_len:	length of @val
+ *
+ * It's important that the order of the records is not changed by the new @key.
+ * This function is not needed to replace an old value with a new one of the
+ * same length: it can just be overwritten in place.
+ *
+ * Returns 0 on success, and @query is left pointing to the same record; returns
+ * a negative error code in case of failure, which may be -EAGAIN if a split
+ * happened and the caller must retry.
+ */
+static int __apfs_btree_replace(struct apfs_query *query, void *key, int key_len, void *val, int val_len)
+{
+	struct apfs_node *node = query->node;
+	struct super_block *sb = node->object.sb;
+	struct apfs_btree_node_phys *node_raw;
+	int needed_room;
+	int err;
+
+	ASSERT(key || val);
+	apfs_assert_query_is_valid(query);
+
+	err = apfs_query_join_transaction(query);
+	if (err) {
+		apfs_err(sb, "query join failed");
+		return err;
+	}
+
+	node = query->node;
+	node_raw = (void *)node->object.data;
+	apfs_assert_in_transaction(sb, &node_raw->btn_o);
+
+	needed_room = key_len + val_len;
+	/* We can reuse the space of the replaced key/value */
+	if (key)
+		needed_room -= query->key_len;
+	if (val)
+		needed_room -= query->len;
+
+	if (!apfs_node_has_room(node, needed_room, true /* replace */)) {
+		if (node->records == 1) {
+			apfs_alert(sb, "no room in empty node?");
+			return -EFSCORRUPTED;
+		}
+		err = apfs_node_split(query);
+		if (err && err != -EAGAIN) {
+			apfs_err(sb, "node split failed");
+			return err;
+		}
+		return -EAGAIN;
+	}
+
+	apfs_assert_query_is_valid(query);
+
+	/* The first key in a node must match the parent record's */
+	if (key && query->parent && query->index == 0) {
+		err = __apfs_btree_replace(query->parent, key, key_len, NULL /* val */, 0 /* val_len */);
+		if (err) {
+			if (err != -EAGAIN)
+				apfs_err(sb, "parent update failed");
+			return err;
+		}
+	}
+
+	apfs_assert_query_is_valid(query);
+
+	err = apfs_node_replace(query, key, key_len, val, val_len);
+	if (err) {
+		apfs_err(sb, "node record replacement failed");
+		return err;
+	}
+	return 0;
+}
+
+/**
+ * apfs_btree_replace - Replace a record in a b-tree leaf
  * @query:	exact query that found the record
  * @key:	new on-disk record key (NULL if unchanged)
  * @key_len:	length of @key
@@ -909,68 +1121,54 @@ int apfs_btree_remove(struct apfs_query *query)
  * Returns 0 on success, and @query is left pointing to the same record; returns
  * a negative error code in case of failure.
  */
-int apfs_btree_replace(struct apfs_query *query, void *key, int key_len,
-		       void *val, int val_len)
+int apfs_btree_replace(struct apfs_query *query, void *key, int key_len, void *val, int val_len)
 {
-	struct apfs_node *node = query->node;
-	struct super_block *sb = node->object.sb;
-	struct apfs_btree_node_phys *node_raw;
+	struct super_block *sb = NULL;
+	struct apfs_node *root = NULL, *leaf = NULL;
 	int err;
 
-	ASSERT(key || val);
+	root = apfs_query_root(query);
+	ASSERT(apfs_node_is_root(root));
+	leaf = query->node;
+	ASSERT(apfs_node_is_leaf(leaf));
+	sb = root->object.sb;
 
-	/* Do this first, or node splits may cause @query->parent to be gone */
-	if (apfs_node_is_leaf(node)) {
-		if (apfs_query_is_orphan(query)) {
-			err = apfs_query_refresh(query);
+	while (true) {
+		err = __apfs_btree_replace(query, key, key_len, val, val_len);
+		if (err != -EAGAIN) {
 			if (err)
 				return err;
+			break;
 		}
-		apfs_btree_change_rec_count(query, 0 /* change */,
-					    key_len, val_len);
-	}
-
-	err = apfs_query_join_transaction(query);
-	if (err) {
-		apfs_err(sb, "query join failed");
-		return err;
-	}
-
-again:
-	node = query->node;
-	node_raw = (void *)node->object.data;
-	apfs_assert_in_transaction(sb, &node_raw->btn_o);
-
-	/* The first key in a node must match the parent record's */
-	if (key && query->parent && query->index == 0) {
-		err = apfs_btree_replace(query->parent, key, key_len,
-					 NULL /* val */, 0 /* val_len */);
+		err = apfs_query_refresh(query, root, false /* nodata */);
 		if (err) {
-			apfs_err(sb, "parent update failed");
+			apfs_err(sb, "query refresh failed");
 			return err;
 		}
 	}
 
-	err = apfs_node_replace(query, key, key_len, val, val_len);
-	if (err == -ENOSPC) {
-		if (!query->parent && !apfs_node_is_root(node)) {
-			if (node->records == 1) {
-				/* Node is defragmented, ENOSPC is absurd */
-				WARN_ON(1);
-				return -EFSCORRUPTED;
-			}
-			err = apfs_query_refresh(query);
-			if (err) {
-				apfs_err(sb, "query refresh failed");
-				return err;
-			}
-		}
-		err = apfs_node_split(query);
-		if (err) {
-			apfs_err(sb, "node split failed");
-			return err;
-		}
-		goto again;
+	apfs_assert_query_is_valid(query);
+	apfs_btree_change_rec_count(query, 0 /* change */, key_len, val_len);
+	return 0;
+}
+
+/**
+ * apfs_query_direct_forward - Set a query to start listing records forwards
+ * @query: a successfully executed query
+ *
+ * Multiple queries list records backwards, but queries marked with this
+ * function after execution will go in the opposite direction.
+ */
+void apfs_query_direct_forward(struct apfs_query *query)
+{
+	if (query->flags & APFS_QUERY_PREV)
+		return;
+
+	apfs_assert_query_is_valid(query);
+	ASSERT(apfs_node_is_leaf(query->node));
+
+	while (query) {
+		query->flags |= APFS_QUERY_PREV;
+		query = query->parent;
 	}
-	return err;
 }
